@@ -54,6 +54,7 @@ getXconsole
 PMS="gnome-power-manager kpowersave xfce4-power-manager"
 PMS="$PMS guidance-power-manager.py dalston-power-applet"
 PMS="$PMS mate-power-manager org_kde_powerdevil"
+PMS="$PMS gsd-power"
 
 # If one of those is running or any of several others,
 if pidof x $PMS > /dev/null ||
@@ -68,7 +69,7 @@ fi
 CURRENT_INIT=`ps -p 1 -o comm=`
 if [ "$CURRENT_INIT" = "openrc-init" ]
 then
-	/sbin/openrc-shutdown -p
+	/sbin/openrc-shutdown -p now
 else
 	/sbin/shutdown -h now "Power button pressed"
 fi
